;;; bijin-tokei.el --- bijin-tokei on Emacs
;;;
;;; Version: 2018/06/03
;;;
;;; Code:

(require 'url)
(require 'deferred)
(require 'image)

(defvar bijin-tokei-last-display-time ""
  "美人時計の最終アップデート時間")

(defvar bijin-tokei-timer nil
  "美人時計の定期実行用のタイマー")

(defvar bijin-tokei-url-prefix
  "http://www.bijint.com/assets/toppict/"
  "美人時計のURLのプレフィクス")

(defcustom bijin-tokei-genre
  "jp"
  "美人時計のジャンル"
  :group 'bijin-tokei)

(defvar bijin-tokei-url-mid
  "/pc/"
  "美人時計のURLの中間部品")

(defvar bijin-tokei-url-postfix ".jpg"
  "美人時計のURLのポストフィックス(拡張子)")

(defcustom bijin-tokei-genre-list (list
                                "jp"
                                "tokyo"
                                "osaka"
                                "nagasaki"
                                "kanazawa"
                                "hiroshima"
                                "kagawa"
                                "tochigi"
                                "fukui"
                                "akita"
                                "gunma"
                                "chiba"
                                "hokkaido"
                                "kobe"
                                "niigata"
                                "kanagawa")
  "美人時計のジャンルのリスト"
  :group 'bijin-tokei)

(defcustom bijin-tokei-shuffle-genre nil
  "ジャンルをシャッフルする"
  :group 'bijin-tokei)

(defvar bijin-tokei-buffername "*bijin-tokei*"
  "美人時計バッファ名")

(defvar bijin-tokei-mode-line-format
  '("    "
    mode-line-frame-identification
    mode-line-buffer-identification))

(defvar bijin-tokei-inhibit-resize nil)
(defvar bijin-tokei-debug nil)

(defvar bijin-tokei-jpg-data nil)
(defvar bijin-tokei-last-window-info nil)
(defvar bijin-tokei-lock-flag nil)
(defvar bijin-tokei-url-access-retry      0)
(defvar bijin-tokei-url-access-retry-max 99)

(defvar bijin-tokei-local-map nil)

(defun bijin-tokei-start (&optional othwin)
  "美人時計の開始"
  (interactive)
  (if (eq (window-system) nil)
      (message "Sorry it can't display in no-window-system")
    (progn
      (setq bijin-tokei-url-access-retry 0)
      (setq bijin-tokei-last-display-time "")
      (setq bijin-tokei-jpg-data nil)
      (setq bijin-tokei-last-window-info nil)
      (setq bijin-tokei-lock-flag nil)
      (setq bijin-tokei-timer (run-with-timer t 1 'bijin-tokei))
      (let ((selected))
        (setq selected (selected-window))
        (if othwin
            (switch-to-buffer-other-window bijin-tokei-buffername)
          (switch-to-buffer bijin-tokei-buffername))
        (setq bijin-tokei-local-map (make-keymap))
        (define-key bijin-tokei-local-map "q"  'bijin-tokei-stop)
        (use-local-map bijin-tokei-local-map)
        (select-window selected)))))

(defun bijin-tokei-start-other-window ()
  "美人時計の開始"
  (interactive)
  (bijin-tokei-start t))

(defun bijin-tokei-stop ()
  "美人時計の停止"
  (interactive)
  (cancel-timer bijin-tokei-timer)
  (kill-buffer bijin-tokei-buffername)
  (setq bijin-tokei-jpg-data nil)
  (setq bijin-tokei-last-window-info nil)
  (setq bijin-tokei-lock-flag nil))

(defun bijin-tokei ()
  "更新の必要のあるものだけ"
  (if (not
       (equal bijin-tokei-last-display-time
              (format-time-string "%H%M" (current-time))))
      (bijin-tokei-update-buffer-async)
    (progn
      (if (and bijin-tokei-jpg-data
               (not bijin-tokei-lock-flag))
          (bijin-tokei-put-image bijin-tokei-jpg-data nil)))
    )
  )

(defun bijin-tokei-update-buffer-async ()
  "美人時計 バッファを更新する"
  (let* ((url (concat bijin-tokei-url-prefix
                      (if bijin-tokei-shuffle-genre
                          (nth (random (length bijin-tokei-genre-list)) bijin-tokei-genre-list)
                        bijin-tokei-genre)
                      bijin-tokei-url-mid
                      (format-time-string "%H%M" (current-time))
                      bijin-tokei-url-postfix)))
    (if bijin-tokei-debug
        (progn (setq url-show-status t)
               (message (format "%s" url)))
      (setq url-show-status nil))
    (if (not bijin-tokei-lock-flag)
        (progn
          (setq bijin-tokei-lock-flag t)
          (deferred:$
            (deferred:timeout
              10000 "deferred canceled url-retrieve."
              (deferred:url-retrieve url))
            (deferred:nextc it
              (lambda (buf)
                (setq bijin-tokei-jpg-data (with-current-buffer buf (buffer-string)))
                (setq bijin-tokei-jpg-data (substring bijin-tokei-jpg-data (+ (string-match "\n\n" bijin-tokei-jpg-data) 2)))
                (kill-buffer buf)
                (bijin-tokei-put-image bijin-tokei-jpg-data t)
                (setq bijin-tokei-url-access-retry 0)
                (setq bijin-tokei-last-display-time
                      (format-time-string "%H%M" (current-time)))
                (setq bijin-tokei-lock-flag nil)))
            (deferred:error it ;
              (lambda (err)
                (progn
                  (if (<= bijin-tokei-url-access-retry bijin-tokei-url-access-retry-max)
                      (progn
                        (setq bijin-tokei-url-access-retry (+ bijin-tokei-url-access-retry 1))
                        (if bijin-tokei-debug (message (format "retrying %s" bijin-tokei-url-access-retry))))
                    (progn
                      (if bijin-tokei-debug (message (format "retry-max %s" bijin-tokei-url-access-retry)))
                      (setq bijin-tokei-url-access-retry 0)
                      (setq bijin-tokei-last-display-time
                            (format-time-string "%H%M" (current-time)))))
                  (setq bijin-tokei-lock-flag nil))
                ))
            ))
      (if bijin-tokei-debug
          (message (format "locking now(%s)" (format-time-string "%D-%H:%M:%S" (current-time)))))
      )
    ))

(defun bijin-tokei-put-image (jpg-data overwrite)
  "美人時計 バッファに画像を表示"
  (with-current-buffer (get-buffer-create bijin-tokei-buffername)
    (let* ((bjwin  (car (get-buffer-window-list bijin-tokei-buffername)))
           (bjsize (window-inside-pixel-edges bjwin))
           (width  (+ (- (nth 2 bjsize) (nth 0 bjsize)) 1))
           (height (+ (- (nth 3 bjsize) (nth 1 bjsize)) 1))
           (old_width  (if bijin-tokei-last-window-info (+ (- (nth 2 bijin-tokei-last-window-info) (nth 0 bijin-tokei-last-window-info)) 1) 0))
           (old_height (if bijin-tokei-last-window-info (+ (- (nth 3 bijin-tokei-last-window-info) (nth 1 bijin-tokei-last-window-info)) 1) 0))
           )
      (if (and bjwin
               (or overwrite
                   (not (and (= width old_width) (= height old_height)))))
        (progn
          (setq cursor-type nil)
          (read-only-mode 0)
          (buffer-disable-undo)
          (erase-buffer)
          (goto-char (point-min))
          (if (and (> width 0) (> height 0))
              (if bijin-tokei-inhibit-resize
                  (insert-image (create-image jpg-data 'jpeg :data))
                (insert-image (bjimg--fit-to-size (create-image jpg-data 'jpeg :data) width height 128))))
          (setq mode-line-format bijin-tokei-mode-line-format)
          (read-only-mode 1)
          (setq bijin-tokei-last-window-info bjsize)))
      (if (not bjwin) (setq bijin-tokei-last-window-info nil))
      ))
  )


(defcustom bijin-tokei-image-sh "sh"
  "shell"
  :group 'bijin-tokei)

(defcustom bijin-tokei-image-convert-direct t
  "画像リサイズコマンドの実行方法を設定する。
nilを指定した場合は shell -cを介してコマンドを実行する。
tを指定した場合は直接コマンドを実行する。
パイプを使用したコマンド列の場合は nilを指定しなくてはならない。"
  :group 'bijin-tokei)

(defvar bijin-tokei-image-conv-topng
  '("convert" "-resize" (format "%sx%s" (nth 1 args) (nth 2 args)) "-" "png:-"))
(defvar bijin-tokei-image-conv-tojpg
  '("convert" "-resize" (format "%sx%s" (nth 1 args) (nth 2 args)) "-" "jpg:-"))
(defvar bijin-tokei-image-conv-togif
  '("convert" "-resize" (format "%sx%s" (nth 1 args) (nth 2 args)) "-" "gif:-"))

;
; その他の例。パイプを使用しているので bijin-tokei-image-convert-directに'nil'をセットしなくてはならない。
;
;(defvar bijin-tokei-image-conv-topng
;  '("djpeg" "-ppm" "|" "pnmscale" "-xysize" (format "%s" (nth 1 args)) (format "%s" (nth 2 args)) "|" "pnmtopng"))
;(defvar bijin-tokei-image-conv-tojpg
;  '("djpeg" "-ppm" "|" "pnmscale" "-xysize" (format "%s" (nth 1 args)) (format "%s" (nth 2 args)) "|" "cjpeg" "-quality" "100"))
;(defvar bijin-tokei-image-conv-togif
;  '("djpeg" "-ppm" "|" "pnmscale" "-xysize" (format "%s" (nth 1 args)) (format "%s" (nth 2 args)) "|" "cjpeg" "-quality" "100" "|" "djpeg" "-gif"))

(defun bjimg--call-convert (image &rest args)
  (let ((spec (cdr image)))
    (with-temp-buffer
      (set-buffer-multibyte nil)
      (let* ((coding-system-for-read  'binary)
             (coding-system-for-write 'binary)
             (expand-args (lambda (lst)
                            (let ((conv-cmd))
                              (dolist (e lst)
                                (if (stringp e) (setq conv-cmd (append conv-cmd (list e)))
                                  (setq conv-cmd (append conv-cmd (list (eval e))))))
                              conv-cmd)))
             (convert-command
              (cond
               ((image-type-available-p 'png)  (funcall expand-args bijin-tokei-image-conv-topng))
               ((image-type-available-p 'jpeg) (funcall expand-args bijin-tokei-image-conv-tojpg))
               ((image-type-available-p 'gif)  (funcall expand-args bijin-tokei-image-conv-togif))
               (t
                (list "convert" "-resize" (format "%sx%s" (nth 1 args) (nth 2 args)) "-" "-")))))

        (if (not bijin-tokei-image-convert-direct)
            (setq convert-command (append (list bijin-tokei-image-sh "-c")
                                          (list (mapconcat 'identity convert-command " ")))))
        (if bijin-tokei-debug (message (format "command : %s" convert-command)))
        (cond
         ((plist-get spec :data)
          (insert (plist-get spec :data))
          ;; stdin to current-buffer
          (unless (eq (apply 'call-process-region
                             (point-min) (point-max)
                             (car convert-command)
                             t (current-buffer) nil
                             (cdr convert-command)
                             ) 0)
            (error "Cannot convert image")))
         ((plist-get spec :file)
          ;; stdin to current-buffer
          (unless (eq (apply 'call-process
                             (car convert-command)
                             (plist-get spec :file) t nil
                             (cdr convert-command)
                             ) 0)
            (error "Cannot convert image")))
         (t
          (error "Not a supported image")))
        (let ((img (create-image (buffer-string) nil t)))
          (plist-put (cdr img) 'bijin-tokei-original-image
                     (or (plist-get (cdr image) 'bijin-tokei-original-image)
                         (copy-sequence image)))
          img)))))

(defun bjimg--zoom (image magnification)
  (let* ((pixels (image-size image t))
         (new (bjimg--call-convert
               image
               "-resize"
               (format "%s" (truncate (* (car pixels) magnification)))
               (format "%s" (truncate (* (cdr pixels) magnification))))))

    ;; clone source image properties
    (when (plist-get (cdr image) :margin)
      (plist-put (cdr new) :margin
                 (plist-get (cdr image) :margin)))
    (when (plist-get (cdr image) :relief)
      (plist-put (cdr new) :relief
                 (plist-get (cdr image) :relief)))
    new))

(defun bjimg--fit-to-size (image width height &optional max)
  "Resize IMAGE with preserving magnification."
  (let* ((pixels (image-size image t))
         (margin (or (plist-get (cdr image) :margin) 0))
         (relief (or (plist-get (cdr image) :relief) 0))
         (mr (+ (* 2 margin) (* 2 relief)))
         (w (+ (car pixels) mr))
         (h (+ (cdr pixels) mr))
         (wr (/ width (ftruncate (float w))))
         (hr (/ height (ftruncate (float h))))
         (magnification (min wr hr max)))
    (bjimg--zoom image magnification)))


(provide 'bijin-tokei)
